@rem = '-*- Perl -*-';
@rem = '
@echo off
ccperl %0 %*
goto endofperl
@rem ';

### ----------------------------------------------------------------------
### | Copyright  Unpublished, Rational Software Corporation. All Rights |
### | Reserved.  This software contains proprietary and confidential     |
### | information of Rational and its suppliers.  Use, disclosure or     |
### | reproduction is prohibited without the prior express written       |
### | consent of Rational Software Corporation.                          |
### ----------------------------------------------------------------------
###
#
# Extract versions from a ClearCase text file delta container file.
#
# The filenames of the extracted version's cleartext are of the form:
#     <elem_id>.<idstr>
#   or
#     <elem_id>.<hex_branch_number>.<hex_version_number>
#
# If possible, the VOB will be consulted to provide useful names:
#     <elem_id> = the leafname of the file element (e.g., "foo.c")
#     <idstr> = the short name for a version (e.g., "/main/mybranch/5"),
#               but "/" substituted with "."
#
# When the VOB doesn't contain know about the element or the branch 
# instance oids the following values are used:
#     <elem_id> = the element's object id (oid)
#     <hex_branch_number> = a container relative hex branch instance number
#
# A element branch map file is also created (branch_map.<elem_id>).  
# This file is useful for determining the element's version tree structure.
# The map contains information of the form:
#  branch <hex_branch_num> (<branch_id>) off of branch <hex_branch_num> version <hex_version_num>
#  version <hex_version_num> on branch <hex_branch_num> (<short_version_name>) timestamp
# E.g.,
#  # Branch map for element 9ceb9cdb.07f711d0.a4fc.08:00:69:02:05:33 (/vobs/foo/bar.c)
#  branch 4 (9ceb9cdb.07f711d0.a4fc.00:01:80:7a:0e:c9) off of branch 3 version 2
#  branch 3 (/main/mybranch) off of branch 0 version 4
#  branch 0 (/main) off of branch 0 version 0
#  # Version map for element bar.c
#  version 2 on branch 3 (/main/mybranch/2) 28-Jan-92.13:33:24
#  version 1 on branch 3 (/main/mybranch/1) 26-Jan-92.18:40:45
#  version 0 on branch 3 (/main/mybranch/0) 26-Jan-92.16:40:28
#

# use strict 'vars';
use strict;

######################################################################

sub usage {
    printf STDERR "Usage: $0 [-M] [-n] [-v]\n";
    printf STDERR "         [-B branch] [-b date-time] [-s date-time]\n";
    printf STDERR "         output-dir-pname\n";
    printf STDERR "         { [-a] vob-storage-pname\n";
    printf STDERR "         | text-delta-container [vob-tag] }\n\n";
    printf STDERR "    -B: Recover versions on only branch #.\n";
    printf STDERR "        (#: hex branch instance number)\n";
    printf STDERR "        Default: all branches. 0 for /main\n";
    printf STDERR "    -b: Recover versions before date-time.\n";
    printf STDERR "        Default: none. (Similar to the time rule).\n";
    printf STDERR "    -s: Recover versions since date-time.\n";
    printf STDERR "        Default: only the LATEST version.\n\n";
    printf STDERR "    -M: Generate the map file only (do not extract cleartext).\n";
    printf STDERR "    -n: Do not resolve names.\n";
    printf STDERR "    -a: Scan all areas. Default: only lost+found areas.\n";
    printf STDERR "    -v: Verbose output.\n";
    exit 2;
}

######################################################################

# Usage:
#      &Getopts('a:bc');  # -a takes arg. -b & -c not.
#                         # Sets $::opt_* as a side effect.
#
sub Getopts {
    my($argumentative) = @_;
    my(@args,$first,$rest,$pos);
    my($errs) = 0;
    local($_);
    local($[) = 0;

    @args = split( / */, $argumentative );
    while(@ARGV && ($_ = $ARGV[0]) =~ /^-(.)(.*)/) {
	($first,$rest) = ($1,$2);
	$pos = index($argumentative,$first);
	if($pos >= $[) {
	    if($args[$pos+1] eq ':') {
		shift(@ARGV);
		if($rest eq '') {
		    ++$errs unless @ARGV;
		    $rest = shift(@ARGV);
		}
		eval "\$::opt_$first = \$rest;";
	    }
	    else {
		eval "\$::opt_$first = 1";
		if($rest eq '') {
		    shift(@ARGV);
		}
		else {
		    $ARGV[0] = "-$rest";
		}
	    }
	}
	else {
	    print STDERR "Unknown option: $first\n";
	    ++$errs;
	    if($rest ne '') {
		$ARGV[0] = "-$rest";
	    }
	    else {
		shift(@ARGV);
	    }
	}
    }
    $errs == 0;
}

######################################################################

# Usage:
#      &basename($path);
#
#  If the scalar context
#    return just the filename from $path.
#  Else
#    return the filename and directory from $path.
#
sub basename {
    my($fullname) = shift(@_);
    my($dirpath,$basename);

    if ($::NT) {
        ($dirpath,$basename) = ($fullname =~ /^(.*[:\\\/])?(.*)/);
        $dirpath .= '.\\' unless $dirpath =~ /\\$/;
    }
    else {	# default to Unix
        ($dirpath,$basename) = ($fullname =~ m;^(.*/)?(.*););
        $dirpath = './' unless $dirpath;
    }

    wantarray ? ($basename,$dirpath) : $basename;
}

# Usage:
#      &joinpath(@path);
#
#  Return the pathname concatenated with the right separator.
#
sub joinpath {
    my($sep) = ($::NT) ? '\\' : '/';
    join($sep, @_);
}

# Usage:
#      &joinleafname(@leaf);
#
#  Return the filename with non-legal charactors substituted with '.'.
#
sub joinleafname {
    my($leaf) = &joinpath(@_);
    if ($::NT) {
        $leaf =~ tr;:\\\/;+++;s;
    }
    else {
        $leaf =~ tr;\\\/;++;s;
    }
    $leaf;
}

# Usage:
#      &date2time($datestring);
#
#  Return the time.
#
sub date2time {
    my($d) = shift(@_);
    my($t) = 0;

    if ($d ne "") {
        my($line);

        open(D2T, "$::date2time $d |")
            || die "Can't execute $::date2time $d: $!\n";
        $line = <D2T>;
        close(D2T);
        ($? == 0) || die "Failed $::date2time $d: $?\n";
        chomp($line);
        $t = hex($line);
    }

    $t;
}

# Usage:
#      &time2date($time);
#
#  Return the date-string.
#
sub time2date {
    my($t) = sprintf "%x", shift(@_);
    my($d);

    if ($t ne "") {
        open(T2D, "$::time2date $t |")
            || die "Can't execute $::time2date $t: $!\n";
        $d = <T2D>;
        close(T2D);
        ($? == 0) || die "Failed $::time2date $t: $?\n";
        chomp($d);
    }

    $d;
}

######################################################################

# Usage:
#      &getproducthome();
#
#  Return ATRIAHOME.
#

use ClearCase::CurrentVersion;

sub getproducthome {

    my($home) = ClearCase::CurrentVersion::SetProductHomeEV;

    ($home eq "") && die "Can't find ATRIAHOME directory!\n";

    $home;
}

######################################################################

# Usage:
#      &is_in_vob();
#
#  Return TRUE if the current directory is in a VOB.
#
sub is_in_vob {
    my($rc) = 0;

    if (!$::opt_n) {
        if (open(CT, "$::cleartool describe \.\@\@ 2>&1 |")) {
            my(@lines) = <CT>;
            close(CT);
            if ($? == 0) {
                $rc = grep(/^directory\s+element/, @lines);
            }
        }

        print STDERR
            "Warning: You are not in a VOB so pathnames won't be printed.\n"
                if ($rc == 0);
    }
    $rc;
}

# Usage:
#      &process_element(\@element, $vobtag)
#
#  IO @element	(element-oid)
#		(element-oid, element-pathname, element-name)
#  IN $vobtag	VOB-Tag
#
#  Resolve the element name.
#
sub process_element {
    my($er) = shift(@_);
    my($vobtag) = shift(@_);
    my($elem_oid) = @$er;
    my($pname) = $elem_oid;
    my($lname) = $elem_oid;

    if (!$::opt_n && $vobtag ne "") {
        if (open(CT, "$::cleartool dump oid:$elem_oid\@$vobtag 2>&1 |")) {
            my(@lines) = <CT>;
            close(CT);
            if ($? == 0) {
                ($pname) = grep { s;^(.*)\@\@$;$1; } @lines;
                chomp($pname);
                $lname = &basename($pname) unless ($pname eq "");
                $pname = $elem_oid if ($pname eq "");
            }
        }
    }

    push(@$er, ($pname, $lname));

    if ($::opt_D) {
        printf STDERR "^E $elem_oid%s\n",
        ($pname =~ m;^[/\\];) ? " $pname $lname" : "";
    }
}

# Usage:
#      &process_branch(\@branch, $vobtag)
#
#  IO @branch	(branch-oid, branch-number, off-branch, off-version)
#		(branch-oid, branch-number, off-branch, off-version, branch-name)
#  IN $vobtag	VOB-Tag
#
#  Resolve the branch name.
#
sub process_branch {
    my($br) = shift(@_);
    my($vobtag) = shift(@_);
    my($branch_oid) = @$br;
    my($name) = $branch_oid;

    if (!$::opt_n && $vobtag ne "") {
        if (open(CT, "$::cleartool dump oid:$branch_oid\@$vobtag 2>&1 |")) {
            my(@lines) = <CT>;
            close(CT);
            # Do not check the returned status because it fails if we are
            # not in the view or the VOB is not mounted.
            # die "Failed $::cleartool dump oid:$branch_oid\@$vobtag: $?\n"
            # if ($? != 0);

            ($name) = grep { s;^.*\s+pname=\"([^\"]+)\".*$;$1; } @lines;
            chomp($name);
            $name = $branch_oid if ($name eq "");
        }
    }

    push(@$br, $name);
}

# Usage:
#      &process_branches(\*MAP, \@branches, $vobtag, \@element);
#
#  IN MAP	output filehandle.
#  IO @branches	((branch-oid, ...)
#		 (branch-oid, ...)
#			:
#		 (branch-oid, ...))
#  IN $vobtag	VOB-Tag
#  IN @element	(element-oid, element-pathname, element-name)
#
#  Print the branch map.
#
sub process_branches {
    my($ofh) = shift(@_);
    my($brs) = shift(@_);
    my($vobtag) = shift(@_);
    my($er) = shift(@_);
    my($elem_oid, $elem_pname, $elem_lname) = @$er;
    my($bref) = [];

    for $bref (@$brs) {
        &process_branch($bref, $vobtag);
    }

    printf $ofh "# Branch map for element $elem_oid%s\n",
    ($elem_oid ne $elem_pname)? " ($elem_pname)" : "";

    for $bref (@$brs) {
        my($branch_oid, $branch_num, $off_branch, $off_version, $branch_name)
            = @$bref;

        if ($::opt_D) {
            printf STDERR "^B %s %x %x %x%s\n",
            $branch_oid, $branch_num, $off_branch, $off_version,
            ($branch_name =~ m;^[/\\];) ? " $branch_name" : "";
        }

        printf $ofh "branch %x (%s) off of branch %x version %x\n",
            $branch_num, $branch_name, $off_branch, $off_version;
    }
}

# Usage:
#      &process_version(\@version, $vobtag)
#
#  IO @version	(version-oid, branch-number, version-number, dtm)
#		(version-oid, branch-number, version-number, dtm, idstr)
#  IN $vobtag	VOB-Tag
#
#  Resolve the version name.
#
sub process_version {
    my($vr) = shift(@_);
    my($vobtag) = shift(@_);
    my($ver_oid, $branch_num, $gen_num) = @$vr;
    my($name) = sprintf "%x.%x", $branch_num, $gen_num;

    if (!$::opt_n && $vobtag ne "") {
        if (open(CT, "$::cleartool dump oid:$ver_oid\@$vobtag 2>&1 |")) {
            my($idstr);
            my(@lines) = <CT>;
            close(CT);
            # Do not check the returned status because it fails if we are
            # not in the view or the VOB is not mounted.
            # die "Failed $::cleartool dump oid:$branch_oid\@$vobtag: $?\n"
            # if ($? != 0);

            ($idstr) = grep { s;^idstr=\"([^\"]+)\".*$;$1; } @lines;
            chomp($idstr);
            $name = $idstr unless ($idstr eq "");
        }
    }

    push(@$vr, $name);
}

# Usage:
#      &extract_version(\@version, $tfd, $elem_pname, $elem_lname, $to)
#
#  IO @version	(version-oid, branch-number, version-number, dtm, idstr)
#  IN $tfd	text_file_delta container pathname
#  IN $elem_pname	element pathname
#  IN $elem_lname	element name
#  IN $to	output directory
#
#  Extract the version.
#
sub extract_version {
    my($vr) = shift(@_);
    my($tfd_path) = shift(@_);
    my($elem_pname) = shift(@_);
    my($elem_lname) = shift(@_);
    my($to) = shift(@_);
    my($ver_oid, $branch_num, $gen_num, $dtm, $idstr) = @$vr;
    my($PID);
    my($dest) = &joinpath($to, &joinleafname($elem_lname, $idstr));
    my($typemgr) = 'text_file_delta';
    my($cmd) = "$::invoke_mgr $typemgr construct_version $tfd_path $dest $ver_oid";
    my($rc);
    my($now);

    printf STDOUT "  Fetching %s to %s\n",
    $elem_pname,  $dest;
    printf STDOUT "    [from text_file_delta container %s]\n", $tfd_path;

    truncate($dest, 0) if (-f $dest);
    system($cmd);
    $rc = $?;
    if ($rc != $::MGR_CONSTRUCT_USE_NEW_FILE) {
        print STDERR "Error: Unable to extract a version\n";
        print STDERR "Error: \"$cmd\" returned $rc\n";
        $::numbers{'Not-extracted'}++;
        return;
    }

    if ($::NT) {
        # Well, utime() doesn't work with the pre ver 5.003 Perl.
        # Ver 5.00307 requires the delay.
        sleep(1) if ($] > 5.003) 
    }

    $now = time;
    utime($now, $dtm, $dest);

    $::numbers{'Extracted'}++ if (-f $dest);
}

# Usage:
#      &process_versions(\*MAP, \@versions, $vobtag, \@element, $tfd, $to);
#
#  IN MAP	output filehandle.
#  IO @versions	((version-oid, ...)
#		 (version-oid, ...)
#			:
#		 (version-oid, ...))
#  IN $vobtag	VOB-Tag
#  IN @element	(element-oid, element-pathname, element-name)
#  IN $tfd	text_file_delta container pathname
#  IN $to	output directory
#
#  Print the version map and extract the versions.
#
sub process_versions {
    my($ofh) = shift(@_);
    my($vers) = shift(@_);
    my($vobtag) = shift(@_);
    my($er) = shift(@_);
    my($tfd_path) = shift(@_);
    my($to) = shift(@_);
    my($elem_oid, $elem_pname, $elem_lname) = @$er;
    my($vref) = [];

    for $vref (@$vers) {
        &process_version($vref, $vobtag);
    }

    printf $ofh "# Version map for element $elem_lname\n" if (@$vers > 0);

    for $vref (@$vers) {
        my($ver_oid, $branch_num, $gen_num, $dtm, $idstr) = @$vref;

        if ($::opt_D) {
            printf STDERR "^V %s %x %x %x%s\n",
            $ver_oid, $branch_num, $gen_num, $dtm,
            ($idstr =~ m;^[/\\];) ? " $idstr" : "";
        }

        printf $ofh "version %x on branch %x (%s) %s\n",
        $gen_num, $branch_num, $idstr, &time2date($dtm);

        &extract_version($vref, $tfd_path, $elem_pname, $elem_lname, $to)
            unless ($::opt_M);
    }
}

# Usage:
#      &process_tfd($to, $tfd, $vobtag, $since, $before, $only_branch);
#
#  IN $to	output directory
#  IN $tfd	text_file_delta container pathname
#  IN $vobtag	VOB-Tag
#  IN $since	time
#  IN $before	time
#  IN $only_branch	branch number
#
sub process_tfd {
    my($to) = shift(@_);
    my($tfd_path) = shift(@_);
    my($vobtag) = shift(@_);
    my($since) = shift(@_);
    my($before) = shift(@_);
    my($only_branch) = shift(@_);
    my($only_latest) = ($since == -1);
    my($line);
    my($elem_oid);
    my($elem_pname);
    my($elem_lname);
    my(@versions);
    my(@branches);
    my($mapfile);
    my($eref) = [];
    my($vref) = [];

    if (&basename($tfd_path) =~ m;^tmp_;) {
        $::numbers{'Temporary'}++;
        return;	# skip unstable temp containers
    }

    unless (open(TFD, $tfd_path)) {
        print STDERR "Can't open $tfd_path: $!\n";
        $::numbers{'Unopened'}++;
        return;
    }

    binmode(TFD);

    while (defined($line = <TFD>)) {
        chomp($line);

        # ^V b38bcdc6.208911d1.a94d.08:00:09:41:e4:52 0 0 3406f36f
        if ((!$only_latest || @versions == 0)
            && $line =~ m;^\^V ([\da-f:\.]+)\s+([\da-f]+)\s+([\da-f]+)\s+([\da-f]+)$;) {
            my($ver_oid) = $1;
            my($branch_num) = hex($2);
            my($gen_num) = hex($3);
            my($dtm) = hex($4);

	    # Do only a specific branch?
            if ($only_branch != -1 && $only_branch != $branch_num) {
                next;	# not an interesting one
            }

	    # Do since and before a date?
	    if ($dtm < $since || ($before != 0 && $dtm >= $before)) {
                next;	# not an interesting one
            }

            push(@versions, [ $ver_oid, $branch_num, $gen_num, $dtm ]);
            next;
        }

        # ^B b38bcdc2.208911d1.a94d.08:00:09:41:e4:52 0 0 0
        if ($line =~ m;^\^B ([\da-f:\.]+)\s+([\da-f]+)\s+([\da-f]+)\s+([\da-f]+)$;) {
            my($branch_oid) = $1;
            my($branch_num) = hex($2);
            my($off_branch) = hex($3);
            my($off_version) = hex($4);

            push(@branches,
                 [ $branch_oid, $branch_num, $off_branch, $off_version]);
            next;
        }

        # ^E b38bcdbe.208911d1.a94d.08:00:09:41:e4:52
        if ($line =~ m;^\^E ([\da-f:\.]+)$;) {
            $elem_oid = $1;
            last;
        }
    }

    close(TFD);

    if ($elem_oid eq "") {
        print STDERR "Non text_file_delta container: $tfd_path\n";
        $::numbers{'Non-TFD'}++;
        return;
    }

    printf STDOUT "Processing $tfd_path ...\n" if ($::opt_v);

    $eref = [ $elem_oid ];
    &process_element($eref, $vobtag);
    ($elem_oid, $elem_pname, $elem_lname) = @$eref;

    $mapfile = &joinpath($to, "branch_map.$elem_lname");

    if ($::opt_M || @versions > 0) {
	# Note: only append to map file. That can result in extra
	# swill but in the case where there are multiple containers
	# with the same element name we want to be sure we capture all
	# of the data possible.
        open(MAP, ">>$mapfile") || die "Can't open $mapfile: $!\n";
        print MAP "\n# Path: $tfd_path\n";
        &process_branches(\*MAP, \@branches, $vobtag, $eref);
        &process_versions(\*MAP, \@versions, $vobtag, $eref, $tfd_path, $to);
        close(MAP);
        print STDOUT "Branch map for $elem_pname in $mapfile\n";
        $::numbers{'Mapped'}++;
    }
    else {
        $::numbers{'Unmapped'}++;
    }
}

# Usage:
#      &find_laf($top);
#
#  IN $top	VOB source container pool directory
#
#  Return the list of lost+found directories.
#
sub find_laf {
    my($top) = shift(@_);
    my(@dirs);

    if ($::NT) {
        my($line);

        open(CMD, "$ENV{'ComSpec'} /c dir /s /b /ad $top 2>&1 |")
            || die "Can't execute $ENV{'ComSpec'} /c dir /s /b /ad $top: $!\n";
        while (defined($line = <CMD>)) {
            push(@dirs, $1)
                if ($line =~ m;^(\Q$top\E\\[^:\\\/]+\\lost\+found)$;);
        }
        close(CMD);
        ($? == 0) || die "Failed $ENV{'ComSpec'} /c dir /s /b /ad $top: $!\n";

        chomp(@dirs);
    }
    else {
        @dirs = glob("$top/*/lost+found");
        chomp(@dirs);
    }

    @dirs;
}

# Usage:
#      &process_stg($to, $tfd, $only_laf, $since, $before, $only_branch);
#
#  IN $to	output directory
#  IN $stg	VOB storage root
#  IN $only_laf	True to process only lost+found directories
#  IN $since	time
#  IN $before	time
#  IN $only_branch	branch number
#
sub process_stg {
    my($to) = shift(@_);
    my($stg) = shift(@_);
    my($only_laf) = shift(@_);
    my($since) = shift(@_);
    my($before) = shift(@_);
    my($only_branch) = shift(@_);
    my($sc) = &joinpath($stg, 's');
    my($vobtag);
    my(@dirs) = ($sc);
    my($d);

    (-d $sc && -r $sc && -x $sc)
        || die "$stg doesn't look like a VOB storage root!\n";

    if (!$::opt_n) {
        if (open(CT, "$::cleartool lsvob -l -storage $stg 2>&1 |")) {
            my(@lines) = <CT>;
            close(CT);
            if ($? == 0) {
                ($vobtag) = grep { s;^Tag:\s+([^\s]+).*$;$1; } @lines;
                chomp($vobtag);
            }
        }

        if ($vobtag eq "") {
            print STDERR "Warning: Unable to determine VOB tag.\n";
        }
        else {
            print STDOUT "Using VOB tag: $vobtag\n";
        }
    }
    if ($only_laf) {
        @dirs = &find_laf($sc);
        (@dirs > 0) || die "There is no lost+found areas to scan.\n";
    }

    foreach $d (@dirs) {
        my($file);
        my($cmd) = "find $d -type f -follow -print 2>&1 |";

        $cmd = "$ENV{'ComSpec'} /c dir /s /b /a-d $d 2>&1 |" if ($::NT);

        open(CMD, "$cmd")
            || die "Can't execute $cmd: $!\n";
        while (defined($file = <CMD>)) {
            chomp($file);

            &process_tfd($to, $file, $vobtag, $since, $before, $only_branch);
        }
        close(CMD);
        ($? == 0) || print STDERR "Warning: $cmd failed: $!\n";
    }

    printf STDOUT "\n=====  Summmary  =====\n";
    printf STDOUT "# of skipped files:           %10d\n",
        ($::numbers{'Temporary'} + $::numbers{'Unopened'}
         + $::numbers{'Non-TFD'});
    printf STDOUT "# of uninteresting containers: %10d\n",
        $::numbers{'Unmapped'};
    printf STDOUT "# of processed containers:    %10d\n",
        $::numbers{'Mapped'};
    printf STDOUT "# of extracted versions:      %10d\n",
        $::numbers{'Extracted'};
    printf STDOUT "# of failed extractions:      %10d\n",
        $::numbers{'Not-extracted'}++;
}

######################################################################

# From /usr/atria/lib/mgrs/mgr_info.h
local($::MGR_CONSTRUCT_USE_SRC_CONTAINER) = 101 << 8;
local($::MGR_CONSTRUCT_USE_NEW_FILE) = 102 << 8;

local($::NT);

local($::date2time);
local($::time2date);
local($::cleartool);
local($::invoke_mgr);

local($::opt_s);
local($::opt_b);
local($::opt_B);
local($::opt_M);
local($::opt_n);
local($::opt_a);
local($::opt_v);
local($::opt_D);

local(%::numbers);

my($atriahome);
my($to_dir_pn);
my($stg);
my($vobtag);
my($only_branch) = -1;
my($since_time) = -1;
my($before_time) = 0;

######################################################################

$| = 1;

&Getopts('s:b:B:nDMav') || &usage();

if ($^O eq "") {
    if (-d '\\\\.\\C:\\') {
        $::NT = 1;
    }
    else {
        $::NT = 0;
    }
}
elsif ($^O eq 'MSWin32') {
    $::NT = 1;
}

$atriahome = $ENV{'ATRIAHOME'};
$atriahome = &getproducthome() if ($atriahome eq "");

if ($::NT) {
    $::invoke_mgr = &joinpath($atriahome, 'etc', 'utils', 'invoke_mgr.exe');
    $::date2time = &joinpath($atriahome, 'etc', 'utils', 'date2time.exe');
    $::time2date = &joinpath($atriahome, 'etc', 'utils', 'time2date.exe');
    $::cleartool = &joinpath($atriahome, 'bin', 'cleartool.exe');

    # Just in case that it's case sensitive...
    my($comspec) = grep { s;^ComSpec$;$ENV{$_};i; } keys(%ENV);
    my($systemroot) = grep { s;^SystemRoot$;$ENV{$_};i; } keys(%ENV);
    my($leaf) = &basename($comspec);
    $ENV{'ComSpec'} = &joinpath($systemroot, 'system32', 'cmd.exe')
        unless ($leaf =~ /cmd.exe/i);
}
else {
    $::invoke_mgr = &joinpath($atriahome, 'etc', 'utils', 'invoke_mgr');
    $::date2time = &joinpath($atriahome, 'etc', 'utils', 'date2time');
    $::time2date = &joinpath($atriahome, 'etc', 'utils', 'time2date');
    $::cleartool = &joinpath($atriahome, 'bin', 'cleartool');
}

(-x $::invoke_mgr) || die "Can't execute $::invoke_mgr\n";
(-x $::date2time) || die "Can't execute $::date2time\n";
(-x $::time2date) || die "Can't execute $::time2date\n";
(-x $::cleartool) || die "Can't execute $::cleartool\n";

if ($::NT) {
    # Quote the path because it may contain spaces.
    $::invoke_mgr = '"' . $::invoke_mgr . '"';
    $::date2time = '"' . $::date2time . '"';
    $::time2date = '"' . $::time2date . '"';
    $::cleartool = '"' . $::cleartool . '"';
}

$since_time = &date2time($::opt_s) if (defined($::opt_s));
$before_time = &date2time($::opt_b);
$only_branch = hex($::opt_B) if (defined($::opt_B));

######################################################################

($to_dir_pn, $stg, $vobtag) = @ARGV;

(-d $to_dir_pn) || &usage();
(-d $stg || -f $stg) || &usage();

&is_in_vob();

# $::opt_M = 1;

printf STDOUT "Will recover versions since %s.\n", &time2date($since_time)
    if ($since_time != -1);
printf STDOUT "Will recover versions before %s.\n", &time2date($before_time)
    if ($before_time);
printf STDOUT "Will recover only branch %x.\n", $only_branch
    if ($only_branch != -1);
printf STDOUT "Will generate only the map file.\n" if ($::opt_M);

if (-f $stg) {
    &process_tfd($to_dir_pn, $stg, $vobtag,
                 $since_time, $before_time, $only_branch);
}
else {
    &process_stg($to_dir_pn, $stg, !$::opt_a,
                 $since_time, $before_time, $only_branch);
}


__END__
:endofperl
